mod proxy_header;

pub use proxy_header::ProxyServerHeader;

use std::sync::Arc;

use axum::{
    async_trait,
    extract::FromRequestParts,
    http::{
        header::{CONTENT_LENGTH, DATE, HOST},
        request::Parts,
    },
    Extension,
};
use chrono::{DateTime, Utc};
use foxchat::{
    fed::{SERVER_HEADER, SIGNATURE_HEADER},
    http::ApiError,
    signature::{parse_date, verify_signature},
    FoxError,
};
use tracing::error;

use crate::{app_state::AppState, model::chat_instance::ChatInstance};

/// A parsed and validated federation signature.
pub struct FoxRequestData {
    pub instance: ChatInstance,
}

#[async_trait]
impl<S> FromRequestParts<S> for FoxRequestData
where
    S: Send + Sync,
{
    type Rejection = ApiError;

    async fn from_request_parts(parts: &mut Parts, state: &S) -> Result<Self, Self::Rejection> {
        let signature = FoxSignatureData::from_request_parts(parts, state).await?;
        let state: Extension<Arc<AppState>> = Extension::from_request_parts(parts, state)
            .await
            .expect("AppState was not added as an extension");

        let instance = ChatInstance::get(state.0.clone(), &signature.domain).await?;
        let public_key = instance.parse_public_key()?;
        if state.config.domain.clone() != signature.host {
            return Err(FoxError::InvalidHeader.into());
        }

        if let Err(e) = verify_signature(
            &public_key,
            signature.signature,
            signature.date,
            &signature.host,
            &signature.request_path,
            signature.content_length,
            None,
        ) {
            error!(
                "Verifying signature from request for {} from {}: {}",
                parts.uri.path(),
                signature.domain,
                e
            );

            return Err(FoxError::InvalidSignature.into());
        }

        Ok(FoxRequestData {
            instance,
        })
    }
}

// An unvalidated federation signature. Use with caution, you should almost always use `FoxRequestData` instead.
pub struct FoxSignatureData {
    pub domain: String,
    pub signature: String,
    pub date: DateTime<Utc>,
    pub host: String,
    pub request_path: String,
    pub content_length: Option<usize>,
}

#[async_trait]
impl<S> FromRequestParts<S> for FoxSignatureData
where
    S: Send + Sync,
{
    type Rejection = ApiError;

    async fn from_request_parts(parts: &mut Parts, _state: &S) -> Result<Self, Self::Rejection> {
        let domain = parts
            .headers
            .get(SERVER_HEADER)
            .ok_or(FoxError::InvalidHeader)?
            .to_str()?
            .to_string();

        let date = parse_date(
            parts
                .headers
                .get(DATE)
                .ok_or(FoxError::InvalidHeader)?
                .to_str()?,
        )?;
        let signature = parts
            .headers
            .get(SIGNATURE_HEADER)
            .ok_or(FoxError::MissingSignature)?
            .to_str()?
            .to_string();
        let host = parts
            .headers
            .get(HOST)
            .ok_or(FoxError::InvalidHeader)?
            .to_str()?
            .to_string();

        let content_length = if let Some(raw_length) = parts.headers.get(CONTENT_LENGTH) {
            Some(raw_length.to_str()?.parse::<usize>()?)
        } else {
            None
        };

        Ok(FoxSignatureData {
            domain,
            date,
            signature,
            host,
            request_path: parts.uri.path().to_string(),
            content_length,
        })
    }
}
