use std::sync::Arc;

use axum::{extract::OriginalUri, http::StatusCode, routing::post, Extension, Json, Router};
use eyre::ContextCompat;
use foxchat::{
    fed::{self, request::ResponseError},
    http::ApiError,
    model::{
        http::{channel::CreateMessageParams, guild::CreateGuildParams},
        Guild, Message,
    },
    FoxError,
};
use serde::{de::DeserializeOwned, Serialize};
use tracing::debug;

use crate::{app_state::AppState, fed::ProxyServerHeader};

use super::auth::AuthUser;

pub fn router() -> Router {
    Router::new()
        .route("/guilds", post(proxy_post::<CreateGuildParams, Guild>))
        .route(
            "/channels/:id/messages",
            post(proxy_post::<CreateMessageParams, Message>),
        )
}

async fn proxy_get<R: Serialize + DeserializeOwned>(
    Extension(state): Extension<Arc<AppState>>,
    ProxyServerHeader(remote): ProxyServerHeader,
    AuthUser(user): AuthUser,
    OriginalUri(original_uri): OriginalUri,
) -> Result<Json<R>, ApiError> {
    let original = original_uri.to_string();
    let proxy_path = original
        .strip_prefix("/_fox/proxy/")
        .wrap_err("invalid url")?;

    let resp = fed::get::<R>(
        &state.private_key,
        &state.config.domain,
        &remote.domain,
        &format!("/_fox/chat/{}", proxy_path),
        Some(user.id),
    )
    .await;

    match resp {
        Ok(r) => match r {
            Some(r) => Ok(Json(r)),
            None => Err(FoxError::ResponseNotOk.into()),
        },
        Err(e) => {
            if let Some(e) = e.downcast_ref::<ResponseError>() {
                match e {
                    ResponseError::JsonError => {
                        return Err(FoxError::ProxyInternalServerError.into())
                    }
                    ResponseError::NotOk {
                        status,
                        code,
                        message,
                    } => {
                        return Err(ApiError {
                            status: StatusCode::from_u16(status.to_owned()).map_err(
                                |_| -> ApiError { FoxError::ProxyInternalServerError.into() },
                            )?,
                            code: code.to_owned(),
                            message: message.to_owned(),
                        })
                    }
                }
            } else {
                tracing::error!("proxying GET request: {}", e);
                return Err(FoxError::ProxyInternalServerError.into());
            }
        }
    }
}

async fn proxy_post<B: Serialize, R: Serialize + DeserializeOwned>(
    Extension(state): Extension<Arc<AppState>>,
    ProxyServerHeader(remote): ProxyServerHeader,
    AuthUser(user): AuthUser,
    OriginalUri(original_uri): OriginalUri,
    Json(body): Json<B>,
) -> Result<Json<R>, ApiError> {
    let original = original_uri.to_string();
    let proxy_path = original
        .strip_prefix("/_fox/proxy/")
        .wrap_err("invalid url")?;
    debug!("Proxying request to {}", proxy_path);

    let resp = fed::post::<B, R>(
        &state.private_key,
        &state.config.domain,
        &remote.domain,
        &format!("/_fox/chat/{}", proxy_path),
        Some(user.id),
        &body,
    )
    .await;

    match resp {
        Ok(r) => match r {
            Some(r) => Ok(Json(r)),
            None => Err(FoxError::ResponseNotOk.into()),
        },
        Err(e) => {
            if let Some(e) = e.downcast_ref::<ResponseError>() {
                match e {
                    ResponseError::JsonError => {
                        return Err(FoxError::ProxyInternalServerError.into())
                    }
                    ResponseError::NotOk {
                        status,
                        code,
                        message,
                    } => {
                        return Err(ApiError {
                            status: StatusCode::from_u16(status.to_owned()).map_err(
                                |_| -> ApiError { FoxError::ProxyInternalServerError.into() },
                            )?,
                            code: code.to_owned(),
                            message: message.to_owned(),
                        })
                    }
                }
            } else {
                tracing::error!("proxying POST request: {}", e);
                return Err(FoxError::ProxyInternalServerError.into());
            }
        }
    }
}
