mod app_state;
mod config;
mod db;
mod http;
mod model;
mod fed;

use crate::{config::Config, model::instance::Instance};
use clap::{Parser, Subcommand};
use color_eyre::eyre::Result;
use std::net::{Ipv4Addr, SocketAddrV4};
use tokio::net::TcpListener;
use tracing::info;

#[derive(Debug, Parser)]
struct Cli {
    #[command(subcommand)]
    command: Option<Command>,
}

#[derive(Debug, Subcommand)]
enum Command {
    Serve,
    Migrate,
}

#[tokio::main]
async fn main() -> Result<()> {
    color_eyre::install()?;

    let config = Config::load()?;
    let args = Cli::parse();

    tracing_subscriber::fmt()
        .with_max_level(config.tracing_level().unwrap_or(tracing::Level::INFO))
        .init();

    match args.command.unwrap_or(Command::Serve) {
        Command::Serve => main_web(config).await,
        Command::Migrate => main_migrate(config).await,
    }
}

async fn main_migrate(config: Config) -> Result<()> {
    info!("Connecting to database");
    let pool = db::init(&config.database_url).await?;
    info!("Migrating database");
    sqlx::migrate!().run(&pool).await?;
    info!("Migrated database");
    Ok(())
}

async fn main_web(config: Config) -> Result<()> {
    info!("Connecting to database");
    let pool = db::init(&config.database_url).await?;

    if config.auto_migrate.unwrap_or(false) {
        info!("Auto-migrate is enabled, migrating database");
        sqlx::migrate!().run(&pool).await?;
        info!("Migrated database");
    }

    info!("Initializing instance data");
    db::init_instance(&pool).await?;
    info!("Initialized instance data!");

    let instance = Instance::get(&pool).await?;
    let port = config.port;
    let app = http::new(pool, config, instance);

    let listener = TcpListener::bind(SocketAddrV4::new(Ipv4Addr::new(127, 0, 0, 1), port)).await?;

    axum::serve(listener, app).await?;

    Ok(())
}
