use std::sync::Arc;

use eyre::{Context, Result};
use foxchat::{
    fed::{self, request::is_valid_domain},
    s2s::http::{HelloRequest, HelloResponse},
    FoxError,
};
use rsa::{pkcs1::DecodeRsaPublicKey, RsaPublicKey};
use serde::{Deserialize, Serialize};
use ulid::Ulid;

use crate::app_state::AppState;

#[derive(Serialize)]
pub struct ChatInstance {
    pub id: String,
    pub domain: String,
    pub base_url: String,
    pub public_key: String,
    pub status: InstanceStatus,
    pub reason: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, sqlx::Type)]
#[sqlx(type_name = "instance_status", rename_all = "lowercase")]
pub enum InstanceStatus {
    Active,
    Suspended,
}

impl ChatInstance {
    pub async fn get(state: Arc<AppState>, domain: &str) -> Result<Self> {
        if !is_valid_domain(&domain) {
            return Err(FoxError::InvalidServer.into());
        }

        if let Some(instance) = sqlx::query_as!(
            Self,
            r#"select id, domain, base_url, public_key,
            status as "status: InstanceStatus", reason
            from chat_instances where domain = $1"#,
            domain
        )
        .fetch_optional(&state.pool)
        .await?
        {
            return Ok(instance);
        }

        let resp: HelloResponse = fed::post(
            &state.private_key,
            &state.config.domain,
            &domain,
            "/_fox/chat/hello",
            None,
            &HelloRequest {
                host: state.config.domain.clone(),
            },
        )
        .await?.expect("");

        if resp.host != domain {
            return Err(FoxError::InvalidServer.into());
        }

        let instance = sqlx::query_as!(
            Self,
            r#"insert into chat_instances
        (id, domain, base_url, public_key) values ($1, $2, $3, $4)
        returning id, domain, base_url, public_key,
        status as "status: InstanceStatus", reason"#,
            Ulid::new().to_string(),
            domain,
            format!("https://{domain}"),
            resp.public_key
        )
        .fetch_one(&state.pool)
        .await?;

        Ok(instance)
    }

    pub fn parse_public_key(&self) -> Result<RsaPublicKey> {
        RsaPublicKey::from_pkcs1_pem(&self.public_key)
            .wrap_err("parsing identity instance public key")
    }
}
